#!/usr/bin/ruby

require 'English'
require 'dotenv'

module Dotenv

  module_function

  def load_without_export(*filenames)
    with(*filenames) do |f|
      ignoring_nonexistent_files do
        Environment.new(f, true)
      end
    end
  end

end

module Aggkit

  class Env

    attr_accessor :project, :name, :env_root, :environment

    def self.list(path = Dir.pwd)
      root = Project.new(path).project_root
      Pathfinder.new(root).each_env
    end

    def self.with_env(env)
      current_env = ENV.to_h
      env.each_pair do |k, v|
        ENV[k.to_s] = v.to_s
      end
      yield(env)
    ensure
      (env.keys.map(&:to_s) - current_env.keys).each do |key|
        ENV.delete(key.to_s)
      end

      current_env.each_pair do |k, v|
        ENV[k.to_s] = v.to_s
      end
    end

    def initialize(path)
      @project = if File.directory?(File.expand_path(path))
        @name = File.basename(File.realpath(File.expand_path(path)))
        Project.new(path)
      else
        @name = path.strip
        Project.new(Dir.pwd)
      end

      unless self.class.list.include?(@name)
        raise "There no env #{@name.inspect} in project: #{project.project_root.inspect}"
      end

      @env_root = File.join(project_root, 'envs', @name)
    end

    def prepare(release: false)
      @environment = prepare_environment(release: release)
    end

    def with_env(&block)
      Env.with_env(environment) & block
    end

    def project_root
      project.project_root
    end

    def core_root
      project.core_root
    end

    def dot_file
      File.join(env_root, '.env')
    end

    def project_dot_file
      File.join(project_root, '.env')
    end

    def core_dot_file
      begin
        File.join(core_root, '.env')
      rescue StandardError
        nil
      end
    end

    def env_files
      [core_dot_file, project_dot_file, dot_file].compact.select{|f| File.exist?(f) }.uniq
    end

    def config_file
      [File.join(env_root, 'config.yml'), File.join(env_root, 'config.yaml')].compact.select{|f| File.exist?(f) }.uniq.first
    end

    def show
      {
        name:         name,
        project_root: project_root,
        core_root:    core_root,
        envfiles:     env_files,
        config_file:  config_file,
        envs:         environment
      }
    end

    private


      def prepare_environment(release: false)
        base_env = {
          'ENV_ROOT'     => env_root,
          'PROJECT_ROOT' => project_root,
          'CORE_ROOT'    => core_root,
          'PATH'         => check_path_variable(env_root)
        }

        dot_env = prepare_dot(base_env)

        compose_env = prepare_compose(dot_env.merge(base_env), release: release)

        dot_env.merge(base_env).merge(compose_env)
      end

      def prepare_dot(env)
        Env.with_env(env) do
          Dotenv.load_without_export(*env_files)
        end
      end

      def prepare_compose(env, release: false)
        compose_files = if env['COMPOSE_FILE']
          env['COMPOSE_FILE']
        elsif File.exist?(File.join(env_root, 'docker-compose.yml'))
          File.join(env_root, 'docker-compose.yml')
        elsif File.exist?(File.join(env_root, 'docker-compose.yaml'))
          File.join(env_root, 'docker-compose.yaml')
        end

        if compose_files
          result_file = File.join(env_root, 'compose-result.yml')
          Env.with_env(env.merge('COMPOSE_FILE' => compose_files)) do
            if release
              `aggmerge #{result_file} --skip-build`
            else
              `aggmerge #{result_file}`
            end
          end
          {
            'COMPOSE_FILE' => result_file
          }
        else
          {}
        end
      end

      def check_path_variable(path)
        if ENV['PATH'][path]
          ENV['PATH']
        else
          "#{ENV['PATH']}:#{path}"
        end
      end

      class Pathfinder


        attr_accessor :path

        def initialize(path)
          @path = File.realpath(File.expand_path(path))
        end

        def each_parent
          current = path.split(File::SEPARATOR)
          Enumerator.new do |y|
            until current.empty?
              y << current.join(File::SEPARATOR)
              current.pop
            end
          end
        end

        def each_env
          Dir.chdir(File.join(path, 'envs')) do
            Dir.glob('**/*').select do |f|
              File.directory? f
            end.select do |f|
              File.exist?(File.join(f, '.env')) || File.exist?(File.join(f, 'docker-compose.yml')) || File.exist?(File.join(f, 'docker-compose.yaml'))
            end.sort
          end
        end


      end

      class Project


        attr_accessor :env_root

        def initialize(path)
          raise "env path #{path} is not directory!" unless File.directory?(File.expand_path(path))

          @env_root = File.realpath(File.expand_path(path))
        end

        def project_root
          @project_root = Pathfinder.new(env_root).each_parent.find do |path|
            root?(path)
          end || (raise "Can't find project root")
        end

        def core_root
          @core_root = Pathfinder.new(project_root).each_parent.find do |path|
            core?(path)
          end
        end

        def root?(path)
          return File.realpath(File.expand_path(path)) if File.directory?("#{path}/.git") ||
            File.exist?("#{path}/base-service.yml") ||
            File.exist?("#{path}/common-services.yml") ||
            File.directory?("#{path}/envs")
        end

        def core?(path)
          return File.realpath(File.expand_path(path)) if File.exist?("#{path}/.core")
        end


      end


  end

end

