require 'English'

class Aggkit::Watcher::ProcessHandler

  attr_accessor :watcher, :command, :options, :process, :stdin, :stdout, :stderr

  def self.capture(cmd)
    io = IO.popen(cmd)
    output = io.read
    begin
      io.close
    rescue StandardError
      nil
    end
    [output, $?]
  end

  def initialize(watcher, *cmd)
    @watcher = watcher

    @options = if cmd.last.is_a? Hash
      cmd.pop
    else
      {}
    end

    @command = cmd.flatten.map{|c| c.to_s.strip }.reject(&:empty?)
    @process = build_process(*command)

    initialize_streams

    @process.start
  end

  def initialize_streams
    @threads = []

    @threads << Thread.new do
      loop do
        break unless synchro_readline(stdout, STDOUT)
      end
    end

    @threads << Thread.new do
      loop do
        break unless synchro_readline(stderr, STDERR)
      end
    end
  end

  def synchro_readline(io, out)
    str = io.gets
    @watcher.iolock.synchronize{ out.puts(str) }
    true
  rescue StandardError
    false
  end

  def build_process(*cmd)
    pr = ::Aggkit::ChildProcess.build(*cmd)
    @stdout, wout = IO.pipe
    @stderr, werr = IO.pipe

    @stdin = pr.io.stdin

    pr.io.stdout = wout
    pr.io.stderr = werr
    pr.duplex = true
    pr
  end

  def handled?
    !!@process.exit_code
  end

  def stop!(status)
    @process.send(:set_exit_code, status)
    stop
  end

  def terminate
    return if exited?
    return if @terminating

    @terminating = true

    if @options[:termcmd]
      termcmd = @options[:termcmd].gsub('%PID%', pid.to_s)
      @watcher.log "Terminating by #{termcmd}..."
      output, status = ::Aggkit::Watcher::ProcessHandler.capture(termcmd)

      if status.success?
        @watcher.log "Success: #{output}"
      else
        @watcher.error "Failed: #{output}"
        @process.send(:send_term)
      end
    else
      @process.send(:send_term)
    end
  end

  def stop(timeout = (@options[:timeout] || 5))
    return if exited?

    terminate

    begin
      return poll_for_exit(timeout)
    rescue TimeoutError
      # try next
    end

    begin
      @process.send(:send_kill)
    rescue Errno::ECHILD, Errno::ESRCH
      # handle race condition where process dies between timeout
      # and send_kill
    end

    wait
  ensure
    clean_all
  end

  def method_missing(m, *args, &block)
    if @process.respond_to? m
      @process.send(m, *args, &block)
    else
      super
    end
  end

  private

    def clean_all
      @stdin.close rescue nil
      @stdout.close rescue nil
      @stderr.close rescue nil

      @threads.each do |th|
        th.terminate rescue nil
        th.join rescue nil
      end

      @threads = []
    end

end

